from typing import Any


def has_default_input(node: dict, input_name: str, value: Any) -> bool:
    for default_input in node["default_inputs"]:
        if default_input["name"] == input_name:
            return default_input["value"] == value
    else:
        return False



from . import utils
from ewokscore.graph.serialize import GraphRepresentation

_REPRESENTATIONS = [str(s).split(".")[-1] for s in GraphRepresentation]


def add_execute_parameters(parser):
    utils.add_common_parameters(parser)
    utils.add_subworkflows_parameters(parser)
    utils.add_ewoks_inputs_parameters(parser)

    parser.add_argument(
        "--workflow-format",
        type=str.lower,
        default="",
        dest="representation",
        choices=_REPRESENTATIONS,
        help="Source format",
    )
    parser.add_argument(
        "--data-root-uri",
        type=str,
        default="",
        dest="data_root_uri",
        help="Root for saving task results",
    )
    parser.add_argument(
        "--data-scheme",
        type=str,
        choices=["nexus", "json"],
        default="nexus",
        dest="data_scheme",
        help="Default task result format",
    )
    parser.add_argument(
        "-o",
        "--option",
        dest="options",
        action="append",
        default=[],
        metavar="OPTION=VALUE",
        help="Execution option",
    )
    parser.add_argument(
        "-t",
        "--task-option",
        dest="task_options",
        action="append",
        default=[],
        metavar="OPTION=VALUE",
        help="Ewoks task option",
    )
    parser.add_argument(
        "-j",
        "--jobid",
        dest="job_id",
        type=str,
        default=None,
        help="Job id for ewoks events",
    )
    parser.add_argument(
        "--disable-events",
        action="store_true",
        help="Disable ewoks events",
    )
    parser.add_argument(
        "--sqlite3",
        dest="sqlite3_uri",
        type=str,
        default=None,
        help="Store ewoks events in an Sqlite3 database",
    )
    parser.add_argument(
        "--outputs",
        type=str,
        choices=["none", "end", "all"],
        default="none",
        help="Log outputs (per task or merged values dictionary)",
    )
    parser.add_argument(
        "--merge-outputs",
        action="store_true",
        dest="merge_outputs",
        help="Merge node outputs",
    )
    parser.add_argument(
        "--engine",
        type=str,
        choices=["none", "dask", "ppf", "orange"],
        default="none",
        help="Execution engine to be used",
    )


def apply_execute_parameters(args):
    args.workflows, args.graphs = utils.parse_workflows(args)

    inputs = [
        utils.parse_parameter(input_item, args.node_attr, args.inputs == "all")
        for input_item in args.parameters
    ]

    if args.outputs == "all":
        outputs = [{"all": True}]
    elif args.outputs == "end":
        outputs = [{"all": False}]
    else:
        outputs = []

    varinfo = {
        "root_uri": args.data_root_uri,
        "scheme": args.data_scheme,
    }

    load_options = dict()
    if args.root_module:
        load_options["root_module"] = args.root_module
    if args.root_dir:
        load_options["root_dir"] = args.root_dir
    if args.representation:
        load_options["representation"] = args.representation

    execinfo = dict()
    if args.job_id:
        execinfo["job_id"] = args.job_id
    if args.sqlite3_uri:
        # TODO: asynchronous handling may loose events
        execinfo["asynchronous"] = False
        execinfo["handlers"] = [
            {
                "class": "ewokscore.events.handlers.Sqlite3EwoksEventHandler",
                "arguments": [{"name": "uri", "value": args.sqlite3_uri}],
            }
        ]

    task_options = dict(utils.parse_option(item) for item in args.task_options)

    execute_options = dict(utils.parse_option(item) for item in args.options)
    execute_options["inputs"] = inputs
    execute_options["outputs"] = outputs
    execute_options["merge_outputs"] = args.merge_outputs
    execute_options["load_options"] = load_options
    execute_options["varinfo"] = varinfo
    execute_options["execinfo"] = execinfo
    execute_options["task_options"] = task_options

    args.execute_options = execute_options



from . import cli_log_utils
from . import cli_execute_utils
from . import cli_convert_utils
from . import cli_submit_utils
from . import cli_install_utils


def add_execute_parameters(parser, shell=False):
    if shell:
        cli_log_utils.add_log_parameters(parser)
    cli_execute_utils.add_execute_parameters(parser)


def apply_execute_parameters(args, shell=False):
    if shell:
        cli_log_utils.apply_log_parameters(args)
    cli_execute_utils.apply_execute_parameters(args)


def add_convert_parameters(parser, shell=False):
    if shell:
        cli_log_utils.add_log_parameters(parser)
    cli_convert_utils.add_convert_parameters(parser)


def apply_convert_parameters(args, shell=False):
    if shell:
        cli_log_utils.apply_log_parameters(args)
    cli_convert_utils.apply_convert_parameters(args)


def add_submit_parameters(parser, shell=False):
    if shell:
        cli_log_utils.add_log_parameters(parser)
    cli_submit_utils.add_submit_parameters(parser)


def apply_submit_parameters(args, shell=False):
    if shell:
        cli_log_utils.apply_log_parameters(args)
    cli_submit_utils.apply_submit_parameters(args)


def add_install_parameters(parser, shell=False):
    if shell:
        # Show logs of install for better UX
        cli_log_utils.add_log_parameters(parser, default="info")
    cli_install_utils.add_install_parameters(parser)


def apply_install_parameters(args, shell=False):
    if shell:
        cli_log_utils.apply_log_parameters(args)
    cli_install_utils.apply_install_parameters(args)



import os
import logging
import importlib
import sys
from warnings import warn
from typing import Any, Optional, List, Union
from ewokscore.graph import TaskGraph
from ewokscore.events.contexts import job_context, RawExecInfoType


from .cliutils.utils import AbortException
from .cliutils.utils import pip_install
from .utils import extract_requirements, save_current_env_as_requirements
from . import graph_cache

try:
    from ewoksjob.client import submit
except ImportError:
    submit = None

try:
    from pyicat_plus.client.main import IcatClient
    from pyicat_plus.client import defaults as icat_defaults
except ImportError:
    IcatClient = None
    icat_defaults = None


__all__ = ["execute_graph", "load_graph", "save_graph", "convert_graph", "submit_graph"]

logger = logging.getLogger(__name__)


def import_binding(engine: Optional[str]):
    if not engine or engine.lower() == "none":
        binding = "ewokscore"
    elif engine.startswith("ewoks"):
        warn(
            f"engine = '{engine}' is deprecated in favor of '{engine[5:]}'",
            DeprecationWarning,
        )
        binding = engine
    else:
        binding = "ewoks" + engine
    return importlib.import_module(binding)


def execute_graph(
    graph,
    engine: Optional[str] = None,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    execinfo: RawExecInfoType = None,
    environment: Optional[dict] = None,
    convert_destination: Optional[Any] = None,
    save_options: Optional[dict] = None,
    upload_parameters: Optional[dict] = None,
    **execute_options,
):
    with job_context(execinfo, engine=engine) as execinfo:
        if environment:
            environment = {k: str(v) for k, v in environment.items()}
            os.environ.update(environment)

        # Load the graph
        if load_options is None:
            load_options = dict()
        graph = load_graph(graph, inputs=inputs, **load_options)

        # Save the graph (with inputs)
        if convert_destination is not None:
            convert_graph(graph, convert_destination, save_options=save_options)

        # Execute the graph
        mod = import_binding(engine)
        result = mod.execute_graph(graph, execinfo=execinfo, **execute_options)

        # Upload results
        if upload_parameters:
            _upload_result(upload_parameters)
        return result


def _upload_result(upload_parameters):
    if IcatClient is None:
        raise RuntimeError("requires pyicat-plus")
    metadata_urls = upload_parameters.pop(
        "metadata_urls", icat_defaults.METADATA_BROKERS
    )
    client = IcatClient(metadata_urls=metadata_urls)
    logger.info(
        "Sending processed dataset '%s' to ICAT: %s",
        upload_parameters.get("dataset"),
        upload_parameters.get("path"),
    )
    client.store_processed_data(**upload_parameters)


def submit_graph(
    graph,
    _celery_options=None,
    resolve_graph_remotely: Optional[bool] = None,
    load_options: Optional[dict] = None,
    **options,
):
    """Submit a workflow to be executed remotely. The workflow is
    resolved on the client-side by default (e.g. load from a file)
    but can optionally be resolved remotely.
    """
    if submit is None:
        raise RuntimeError("requires the 'ewoksjob' package")
    if _celery_options is None:
        _celery_options = dict()
    if resolve_graph_remotely:
        options["load_options"] = load_options
    else:
        # Do not save requirements since the current env is the client
        graph = convert_graph(
            graph, None, load_options=load_options, save_requirements=False
        )
    return submit(args=(graph,), kwargs=options, **_celery_options)


@graph_cache.cache
def load_graph(
    graph: Any, inputs: Optional[List[dict]] = None, **load_options
) -> TaskGraph:
    """When load option `graph_cache_max_size > 0` is provided, the graph will cached in memory.
    When the graph comes from external storage (for example a file) any changes
    to the external graph will require flushing the cache with `graph_cache_max_size = 0`.
    """
    engine = _get_engine_for_format(graph, options=load_options)
    mod = import_binding(engine)
    return mod.load_graph(graph, inputs=inputs, **load_options)


def save_graph(graph: TaskGraph, destination, **save_options) -> Union[str, dict]:
    engine = _get_engine_for_format(destination, options=save_options)
    mod = import_binding(engine)
    return mod.save_graph(graph, destination, **save_options)


def convert_graph(
    source,
    destination,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    save_options: Optional[dict] = None,
    save_requirements: bool = True,
) -> Union[str, dict]:
    if load_options is None:
        load_options = dict()
    if save_options is None:
        save_options = dict()
    graph = load_graph(source, inputs=inputs, **load_options)
    if save_requirements:
        graph = save_current_env_as_requirements(graph)
    return save_graph(graph, destination, **save_options)


def install_graph(
    source,
    skip_prompt: bool = False,
    python_path: Optional[str] = None,
    load_options: Optional[dict] = None,
):
    if load_options is None:
        load_options = dict()
    graph = load_graph(source, **load_options)

    requirements = graph.requirements
    if requirements is None:
        logger.warning(
            "Requirements field is empty. Trying to extract requirements automatically..."
        )
        requirements = extract_requirements(graph)
        logger.info(f"Extracted the following requirements: {requirements}")

    if python_path is None:
        python_path = sys.executable

    if skip_prompt:
        pip_install(requirements, python_path)
        return

    requirements_as_str = "\n".join(requirements)

    answer = input(
        f"{requirements_as_str}\nThis will install the above packages via {python_path} -m pip install. Do you want to proceed (y/N)?"
    )
    if answer.lower() == "y" or answer.lower() == "yes":
        pip_install(requirements, python_path)
    else:
        raise AbortException()


def _get_engine_for_format(graph, options: Optional[dict] = None) -> Optional[str]:
    """Get the engine which implements the workflow format (loading and saving)."""
    representation = None
    if options:
        representation = options.get("representation")
    if (
        representation is None
        and isinstance(graph, str)
        and graph.lower().endswith(".ows")
    ):
        representation = "ows"
    if representation == "ows":
        return "orange"



from subprocess import CalledProcessError
import sys
import argparse
import traceback
from typing import Optional

from pprint import pprint
from . import cliutils
from .cliutils.utils import AbortException
from .bindings import execute_graph
from .bindings import convert_graph
from .bindings import submit_graph
from .bindings import install_graph


def create_argument_parser(shell=False):
    parser = argparse.ArgumentParser(
        description="Extensible WOrKflow System CLI",
        prog="ewoks",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    subparsers = parser.add_subparsers(help="Commands", dest="command")
    execute = subparsers.add_parser(
        "execute",
        help="Execute a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    submit = subparsers.add_parser(
        "submit",
        help="Schedule a workflow execution",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    convert = subparsers.add_parser(
        "convert",
        help="Convert a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    install = subparsers.add_parser(
        "install",
        help="Install requirements of a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    cliutils.add_execute_parameters(execute, shell=shell)
    cliutils.add_submit_parameters(submit, shell=shell)
    cliutils.add_convert_parameters(convert, shell=shell)
    cliutils.add_install_parameters(install, shell=shell)
    return parser


def command_execute(args, shell=False):
    cliutils.apply_execute_parameters(args, shell=shell)

    return_code = 0
    keep_results = []
    for workflow, graph in zip(args.workflows, args.graphs):
        print("###################################")
        print(f"# Execute workflow '{workflow}'")
        print("###################################")
        try:
            results = execute_graph(graph, engine=args.engine, **args.execute_options)
        except Exception as ex:
            traceback.print_exc()
            print("FAILED")
            results = ex
            return_code = 1
        else:
            if args.outputs == "none":
                if results is None:
                    print("FAILED")
                else:
                    print("FINISHED")
            else:
                print("")
                print("RESULTS:")
                pprint(results)
                print("")
                print("FINISHED")
            if results is None:
                return_code = 1
        finally:
            print()
        if not shell:
            keep_results.append(results)

    if shell:
        return return_code
    return keep_results


def command_submit(args, shell=False):
    cliutils.apply_submit_parameters(args, shell=shell)

    return_code = 0
    keep_results = []

    futures = list()
    for workflow, graph in zip(args.workflows, args.graphs):
        future = submit_graph(
            graph,
            engine=args.engine,
            resolve_graph_remotely=args.resolve_graph_remotely,
            **args.execute_options,
            _celery_options=args.cparameters,
        )
        print(f"Workflow '{workflow}' submitted (ID: {future.task_id})")
        futures.append(future)
    if args.wait < 0:
        if shell:
            return return_code
        return keep_results

    print("Waiting for results ...")
    print()
    for workflow, future in zip(args.workflows, futures):
        print(
            "###########################################################################"
        )
        print(f"# Result of workflow '{workflow}' (ID: {future.task_id})")
        print(
            "###########################################################################"
        )
        try:
            results = future.get(timeout=args.wait)
        except Exception as ex:
            if _is_timeout(ex):
                print(f"Not finished after {args.wait}s")
            else:
                traceback.print_exc()
                print("FAILED")
            results = ex
            return_code = 1
        else:
            if args.outputs == "none":
                if results is None:
                    print("FAILED")
                else:
                    print("FINISHED")
            else:
                pprint(results)
                print("FINISHED")
            if results is None:
                return_code = 1
        finally:
            print()
        if not shell:
            keep_results.append(results)

    if shell:
        return return_code
    return keep_results


def _is_timeout(exception: Optional[Exception]) -> bool:
    if exception is None:
        return False
    if isinstance(exception, TimeoutError):
        return True
    if _is_timeout(exception.__cause__):
        return True
    if _is_timeout(exception.__context__):
        return True
    return False


def command_convert(args, shell=False):
    cliutils.apply_convert_parameters(args, shell=shell)
    for workflow, graph, destination in zip(
        args.workflows, args.graphs, args.destinations
    ):
        convert_graph(graph, destination, **args.convert_options)
        print(f"Converted {workflow} -> {destination}")


def command_install(args, shell=False):
    cliutils.apply_install_parameters(args, shell=shell)
    for workflow, graph in zip(args.workflows, args.graphs):
        try:
            install_graph(graph, args.yes, args.python)
        except CalledProcessError as e:
            print(f"Install failed for {workflow}: {e}")
        except AbortException:
            print(f"Install aborted for {workflow}")
        else:
            print(f"Installed requirements for {workflow}")


def command_default(args, shell=False):
    if shell:
        return 0
    return None


def main(argv=None, shell=True):
    parser = create_argument_parser(shell=shell)

    if argv is None:
        argv = sys.argv
    args = parser.parse_args(argv[1:])

    if args.command == "execute":
        return command_execute(args, shell=shell)
    elif args.command == "submit":
        return command_submit(args, shell=shell)
    elif args.command == "convert":
        return command_convert(args, shell=shell)
    elif args.command == "install":
        return command_install(args, shell=shell)
    else:
        parser.print_help()
        return command_default(args, shell=shell)


if __name__ == "__main__":
    sys.exit(main())



from argparse import ArgumentParser
import os
import json
import subprocess
from glob import glob
from fnmatch import fnmatch
from typing import Sequence, Tuple, Any, List
from json.decoder import JSONDecodeError


class AbortException(Exception):
    pass


def add_common_parameters(parser: ArgumentParser):
    parser.add_argument(
        "workflows",
        type=str,
        help="Workflow to execute (e.g. JSON filename)",
        nargs="+",
    )
    parser.add_argument(
        "--test",
        action="store_true",
        help="The 'workflow' argument refers to the name of a test graph",
    )
    parser.add_argument(
        "--search",
        action="store_true",
        help="The 'workflow' argument is a pattern to be search",
    )


def add_subworkflows_parameters(parser: ArgumentParser):
    parser.add_argument(
        "--workflow-dir",
        type=str,
        default="",
        dest="root_dir",
        help="Directory of sub-workflows (current working directory by default)",
    )
    parser.add_argument(
        "--workflow-module",
        type=str,
        default="",
        dest="root_module",
        help="Python module of sub-workflows (current working directory by default)",
    )


def add_ewoks_inputs_parameters(parser: ArgumentParser):
    parser.add_argument(
        "-p",
        "--parameter",
        dest="parameters",
        action="append",
        default=[],
        metavar="[NODE:]NAME=VALUE",
        help="Input variable for a particular node (or all start nodes when missing)",
    )
    parser.add_argument(
        "--input-node-id",
        dest="node_attr",
        type=str,
        choices=["id", "label", "taskid"],
        default="id",
        help="The NODE attribute used when specifying an input parameter with [NODE:]NAME=VALUE",
    )
    parser.add_argument(
        "--inputs",
        type=str,
        choices=["start", "all"],
        default="start",
        help="Inputs without a specific node are given to either all start nodes or all nodes",
    )


def parse_ewoks_inputs_parameters(args):
    return [
        parse_parameter(input_item, args.node_attr, args.inputs == "all")
        for input_item in args.parameters
    ]


def parse_value(value: str) -> Any:
    try:
        return json.loads(value)
    except JSONDecodeError:
        return value


_NODE_ATTR_MAP = {"id": "id", "label": "label", "taskid": "task_identifier"}


def parse_parameter(input_item: str, node_attr: str, all: bool) -> dict:
    """The format of `input_item` is `"[NODE]:name=value"`"""
    node_and_name, _, value = input_item.partition("=")
    a, sep, b = node_and_name.partition(":")
    if sep:
        node = a
        var_name = b
    else:
        node = None
        var_name = a
    var_value = parse_value(value)
    if node is None:
        return {"all": all, "name": var_name, "value": var_value}
    return {
        _NODE_ATTR_MAP[node_attr]: node,
        "name": var_name,
        "value": var_value,
    }


def parse_option(option: str) -> Tuple[str, Any]:
    option, _, value = option.partition("=")
    return option, parse_value(value)


def parse_workflows(args) -> Tuple[List[str], List[str]]:
    """
    :returns: workflows (possibly expanded due the search), graphs (execute graph arguments)
    """
    if args.test:
        return _parse_test_workflows(args.workflows, args.search)
    return _parse_workflows(args.workflows, args.search)


def _parse_workflows(workflows: List[str], search: bool) -> Tuple[List[str], List[str]]:
    """
    :returns: workflows (possibly expanded due the search), graphs (execute graph arguments)
    """
    if not search:
        return workflows, workflows
    parsed_workflows = list()
    files = (filename for workflow in workflows for filename in glob(workflow))
    for filename in sorted(files, key=os.path.getmtime):
        if filename not in parsed_workflows:
            parsed_workflows.append(filename)
    return parsed_workflows, parsed_workflows


def _parse_test_workflows(
    workflows: List[str], search: bool
) -> Tuple[List[str], List[dict]]:
    """
    :returns: workflows (possibly expanded due the search), graphs (execute graph arguments)
    """
    from ewokscore.tests.examples.graphs import graph_names, get_graph

    test_workflows = list(graph_names())
    if search:
        parsed_workflows = list()
        for workflow in workflows:
            for test_graph in test_workflows:
                if fnmatch(test_graph, workflow) and test_graph not in parsed_workflows:
                    parsed_workflows.append(test_graph)
    else:
        for workflow in workflows:
            if workflow not in test_workflows:
                raise RuntimeError(
                    f"Test graph '{workflow}' does not exist: {test_workflows}"
                )
        parsed_workflows = workflows

    graphs = [get_graph(workflow)[0] for workflow in parsed_workflows]
    return parsed_workflows, graphs


def parse_destinations(args):
    dest_dirname = os.path.dirname(args.destination)
    basename = os.path.basename(args.destination)
    dest_basename, dest_ext = os.path.splitext(basename)
    if not dest_ext:
        dest_ext = dest_basename
        dest_basename = ""
        if not dest_ext.startswith("."):
            dest_ext = f".{dest_ext}"

    if len(args.workflows) == 1 and dest_basename:
        return [os.path.join(dest_dirname, f"{dest_basename}{dest_ext}")]

    destinations = list()
    for workflow in args.workflows:
        basename, _ = os.path.splitext(os.path.basename(workflow))
        destination = os.path.join(dest_dirname, f"{basename}{dest_basename}{dest_ext}")
        destinations.append(destination)

    return destinations


def pip_install(requirements: Sequence[str], python_path: str) -> int:
    # https://pip.pypa.io/en/stable/user_guide/#using-pip-from-your-program
    return subprocess.check_call([python_path, "-m", "pip", "install", *requirements])



from argparse import ArgumentParser
from . import utils


def add_install_parameters(parser: ArgumentParser):
    utils.add_common_parameters(parser)
    parser.add_argument(
        "--yes",
        action="store_true",
        help="Accept automatically install prompts",
    )
    parser.add_argument(
        "-p",
        "--python",
        type=str,
        help="Python of the env where the packages should be installed. Default: current env Python.",
    )


def apply_install_parameters(args):
    args.workflows, args.graphs = utils.parse_workflows(args)



import os
import sys
import pytest

from ewoks.__main__ import main
from ewokscore import load_graph
from ewokscore.graph import TaskGraph
from ewokscore.tests.examples.graphs import graph_names

from ewoks.tests.utils import has_default_input


def _ewokscore_in_graph_requirements(graph: TaskGraph) -> bool:
    ewokscore_in_req = False
    for requirement in graph.graph.graph["requirements"]:
        if "ewokscore" in requirement:
            ewokscore_in_req = True
            break

    return ewokscore_in_req


@pytest.mark.parametrize("graph_name", graph_names())
def test_convert(graph_name, tmpdir):
    destination = str(tmpdir / f"{graph_name}.json")
    argv = [
        sys.executable,
        "convert",
        graph_name,
        destination,
        "--test",
        "-s",
        "indent=2",
    ]
    main(argv=argv, shell=False)
    assert os.path.exists(destination)

    graph = load_graph(destination)
    assert graph.graph.graph["requirements"] is not None
    assert _ewokscore_in_graph_requirements(graph)


def test_convert_with_all_inputs(tmpdir):
    graph_name = "demo"
    destination = str(tmpdir / f"{graph_name}.json")

    argv = [
        sys.executable,
        "convert",
        graph_name,
        destination,
        "--test",
        "-p",
        "value=10",
        "--inputs=all",
    ]

    main(argv=argv, shell=False)
    assert os.path.exists(destination)

    graph = load_graph(destination)
    for node in graph.graph.nodes.values():
        assert has_default_input(node, "value", 10)


def test_convert_with_taskid_inputs(tmpdir):
    graph_name = "demo"
    destination = str(tmpdir / f"{graph_name}.json")
    taskid = "ewokscore.tests.examples.tasks.sumtask.SumTask"

    argv = [
        sys.executable,
        "convert",
        graph_name,
        destination,
        "--test",
        "-p",
        f"{taskid}:value=test",
        "--input-node-id",
        "taskid",
    ]

    main(argv=argv, shell=False)
    assert os.path.exists(destination)

    graph = load_graph(destination)
    for node in graph.graph.nodes.values():
        if node["task_identifier"] == taskid:
            assert has_default_input(node, "value", "test")



from . import utils
from ewokscore.graph.serialize import GraphRepresentation

_REPRESENTATIONS = [str(s).split(".")[-1] for s in GraphRepresentation]


def add_convert_parameters(parser):
    utils.add_common_parameters(parser)
    utils.add_subworkflows_parameters(parser)
    utils.add_ewoks_inputs_parameters(parser)

    parser.add_argument(
        "destination",
        type=str,
        help="Destination of the conversion (e.g. JSON filename)",
    )
    parser.add_argument(
        "--src-format",
        type=str.lower,
        default="",
        dest="source_representation",
        choices=_REPRESENTATIONS,
        help="Source format",
    )
    parser.add_argument(
        "--dst-format",
        type=str.lower,
        default="",
        dest="destination_representation",
        choices=_REPRESENTATIONS,
        help="Destination format",
    )
    parser.add_argument(
        "-o",
        "--load-option",
        dest="load_options",
        action="append",
        default=[],
        metavar="OPTION=VALUE",
        help="Load options",
    )
    parser.add_argument(
        "-s",
        "--save-option",
        dest="save_options",
        action="append",
        default=[],
        metavar="OPTION=VALUE",
        help="Save options",
    )
    parser.add_argument(
        "--exclude-requirements",
        dest="exclude_requirements",
        action="store_true",
        help="Do not include the packages of the current Python env as requirements in the destination workflow",
    )


def apply_convert_parameters(args):
    args.workflows, args.graphs = utils.parse_workflows(args)
    args.destinations = utils.parse_destinations(args)

    load_options = dict(utils.parse_option(item) for item in args.load_options)
    if args.source_representation:
        load_options["representation"] = args.source_representation
    if args.root_module:
        load_options["root_module"] = args.root_module
    if args.root_dir:
        load_options["root_dir"] = args.root_dir

    save_options = dict(utils.parse_option(item) for item in args.save_options)
    if args.destination_representation:
        save_options["representation"] = args.destination_representation

    convert_options = {
        "save_options": save_options,
        "load_options": load_options,
        "inputs": utils.parse_ewoks_inputs_parameters(args),
    }
    if args.exclude_requirements:
        convert_options["save_requirements"] = False
    args.convert_options = convert_options



from collections import OrderedDict
from functools import wraps
from typing import Dict, Optional

from ewokscore.hashing import uhash
from ewokscore.graph import TaskGraph

_GRAPH_CACHE: Dict[int, TaskGraph] = OrderedDict()
_GRAPH_CACHE_MAX_SIZE: int = 0


def cache(load_method):
    @wraps(wraps)
    def wrapper(*args, graph_cache_max_size: Optional[int] = None, **kw):
        set_cache_max_size(graph_cache_max_size)

        if _GRAPH_CACHE_MAX_SIZE <= 0:
            return load_method(*args, **kw)

        graph_id = uhash((args, kw))
        graph = _GRAPH_CACHE.pop(graph_id, None)
        if graph is None:
            graph = load_method(*args, **kw)
        _GRAPH_CACHE[graph_id] = graph
        _check_cache_size()
        return graph

    return wrapper


def set_cache_max_size(graph_cache_max_size: Optional[int] = None) -> None:
    global _GRAPH_CACHE_MAX_SIZE
    if graph_cache_max_size is not None:
        _GRAPH_CACHE_MAX_SIZE = graph_cache_max_size
        _check_cache_size()


def _check_cache_size() -> None:
    while _GRAPH_CACHE and len(_GRAPH_CACHE) > _GRAPH_CACHE_MAX_SIZE:
        _GRAPH_CACHE.popitem(last=False)



import os
from glob import glob
import pytest
import testbook

root_dir = os.path.join(os.path.dirname(__file__), "notebooks")
notebooks = [os.path.basename(f) for f in glob(os.path.join(root_dir, "*.ipynb"))]


@pytest.mark.parametrize("filename", notebooks)
def test_notebooks(filename):
    if filename == "benchmark.ipynb":
        pytest.skip("benchmark takes too long")
    filename = os.path.join(root_dir, filename)
    with testbook.testbook(filename, execute=True):
        pass



import logging
import subprocess
import sys
from typing import List
from ewokscore.graph import TaskGraph


logger = logging.getLogger(__file__)


def extract_requirements(graph: TaskGraph) -> List[str]:
    imports: set[str] = set()

    for node_id, node in graph.graph.nodes.items():
        task_identifier = node["task_identifier"]
        task_type = node["task_type"]

        if (
            task_type == "class"
            or task_type == "method"
            or task_type == "ppfmethod"
            or task_type == "ppfport"
        ):
            package = task_identifier.split(".")[0]
            if package == "__main__" or "":
                logger.warning(
                    f"Could not extract requirements for node {node_id}: the task identifier is a relative import or a import from __main__."
                )
                continue

            imports.add(package)

        elif task_type == "notebook":
            logger.warning(
                f"Requirement extraction may be incomplete for node {node_id}: {task_type} is only partially supported."
            )
            imports.add("ewokscore[notebook]")

        elif task_type == "script":
            logger.warning(
                f"Requirement extraction cannot be done for scripts (node {node_id})."
            )
        else:
            logger.warning(
                f"Could not extract requirements for node {node_id}: unsupported task type {task_type}."
            )

    return list(imports)


def save_current_env_as_requirements(graph: TaskGraph):

    freeze_output = subprocess.check_output(
        [sys.executable, "-m", "pip", "freeze"], text=True
    )
    requirements = freeze_output.strip().split("\n")

    graph.graph.graph["requirements"] = requirements

    return graph



import json
import argparse

import pytest

from ewoks import cliutils


def test_cli_execute_workflow():
    parser = argparse.ArgumentParser()
    cliutils.add_execute_parameters(parser)
    argv = [
        "acyclic1",
        "acyclic2",
        "--test",
        "-p",
        "a=1",
        "-p",
        "task1:b=test",
        "--workflow-dir",
        "/tmp",
    ]
    args = parser.parse_args(argv)
    cliutils.apply_execute_parameters(args)

    assert len(args.graphs) == 2
    assert args.graphs[0]["graph"]["id"] == "acyclic1"
    assert args.graphs[1]["graph"]["id"] == "acyclic2"

    execute_options = {
        "inputs": [
            {"all": False, "name": "a", "value": 1},
            {"id": "task1", "name": "b", "value": "test"},
        ],
        "merge_outputs": False,
        "outputs": [],
        "task_options": {},
        "varinfo": {"root_uri": "", "scheme": "nexus"},
        "load_options": {"root_dir": "/tmp"},
        "execinfo": {},
    }
    assert args.execute_options == execute_options


def test_cli_convert_workflow():
    parser = argparse.ArgumentParser()
    cliutils.add_convert_parameters(parser)
    argv = [
        "acyclic1",
        "test.json",
        "--test",
        "-p",
        "a=1",
        "-p",
        "task1:b=test",
        "--src-format",
        "yaml",
        "--dst-format",
        "json",
    ]
    args = parser.parse_args(argv)
    cliutils.apply_convert_parameters(args)

    assert len(args.graphs) == 1
    assert args.graphs[0]["graph"]["id"] == "acyclic1"

    convert_options = {
        "inputs": [
            {"all": False, "name": "a", "value": 1},
            {"id": "task1", "name": "b", "value": "test"},
        ],
        "load_options": {"representation": "yaml"},
        "save_options": {"representation": "json"},
    }
    assert args.convert_options == convert_options

    argv = ["acyclic1", "test.json"]
    args = parser.parse_args(argv)
    cliutils.apply_convert_parameters(args)

    assert args.destinations == ["test.json"]

    argv = ["acyclic1", ".json"]
    args = parser.parse_args(argv)
    cliutils.apply_convert_parameters(args)

    assert args.destinations == ["acyclic1.json"]

    argv = ["acyclic1", "json"]
    args = parser.parse_args(argv)
    cliutils.apply_convert_parameters(args)

    assert args.destinations == ["acyclic1.json"]


def test_cli_execute_workflow_search(tmp_path, graph_directory):
    parser = argparse.ArgumentParser()
    cliutils.add_execute_parameters(parser)
    argv = [
        str(tmp_path / "subdir" / "*.json"),
        str(tmp_path / "*.json"),
        "--search",
    ]
    args = parser.parse_args(argv)
    cliutils.apply_execute_parameters(args)

    assert len(args.graphs) == 22
    assert args.graphs == graph_directory


def test_cli_convert_workflow_search(tmp_path, graph_directory):
    parser = argparse.ArgumentParser()
    cliutils.add_convert_parameters(parser)
    argv = [
        str(tmp_path / "subdir" / "*.json"),
        str(tmp_path / "*.json"),
        "_suffix.json",
        "--search",
    ]
    args = parser.parse_args(argv)
    cliutils.apply_convert_parameters(args)

    assert len(args.graphs) == 22
    assert args.graphs == graph_directory


@pytest.fixture()
def graph_directory(tmp_path):
    expected_files = list()

    st_mtime = tmp_path.stat().st_mtime

    for i in range(11, 0, -1):
        filename = tmp_path / f"workflow{i}.json"
        expected_files.append(str(filename))
        st_mtime = _create_workflow(filename, {"graph": {"id": f"graph{i}"}}, st_mtime)

    subdir = tmp_path / "subdir"
    subdir.mkdir(parents=True, exist_ok=True)
    for i in range(11, 0, -1):
        filename = tmp_path / f"sub_workflow{i}.json"
        expected_files.append(str(filename))
        st_mtime = _create_workflow(
            filename, {"graph": {"id": f"sub_graph{i}"}}, st_mtime
        )

    return expected_files


def _create_workflow(filename, content, prev_st_mtime):
    while True:
        with open(filename, "w") as f:
            json.dump(content, f)
        st_mtime = filename.stat().st_mtime
        if st_mtime > prev_st_mtime:
            return st_mtime



import os
import sys
import pytest

from ewoks.__main__ import main
from ewokscore import load_graph
from ewokscore.graph import TaskGraph
from ewokscore.tests.examples.graphs import graph_names
from ewokscore.tests.examples.graphs import get_graph
from ewokscore.tests.utils.results import assert_execute_graph_default_result

from ewoks.tests.utils import has_default_input


def _ewokscore_in_graph_requirements(graph: TaskGraph) -> bool:
    ewokscore_in_req = False
    for requirement in graph.graph.graph["requirements"]:
        if "ewokscore" in requirement:
            ewokscore_in_req = True
            break

    return ewokscore_in_req


@pytest.mark.parametrize("graph_name", graph_names())
@pytest.mark.parametrize("scheme", (None, "json"))
@pytest.mark.parametrize("engine", (None, "dask", "ppf"))
def test_execute(graph_name, scheme, engine, tmpdir):
    if graph_name == "self_trigger":
        pytest.skip(
            "Self-triggering workflow execution is inconsistent: https://gitlab.esrf.fr/workflow/ewoks/ewoksppf/-/issues/16"
        )

    graph, expected = get_graph(graph_name)
    argv = [sys.executable, "execute", graph_name, "--test", "--merge-outputs"]
    if engine:
        argv += ["--engine", engine]
    if engine == "ppf":
        argv += ["--outputs", "end"]
    else:
        argv += ["--outputs", "all"]
    if scheme:
        argv += ["--data-root-uri", str(tmpdir), "--data-scheme", scheme]
        varinfo = {"root_uri": str(tmpdir), "scheme": scheme}
    else:
        varinfo = None

    keep = graph
    ewoksgraph = load_graph(graph)
    non_dag = ewoksgraph.is_cyclic or ewoksgraph.has_conditional_links

    results = main(argv=argv, shell=False)
    assert len(results) == 1

    if non_dag and engine != "ppf":
        assert isinstance(results[0], RuntimeError)
    else:
        assert_execute_graph_default_result(ewoksgraph, results[0], expected, varinfo)
        assert keep == graph


def test_execute_with_convert_destination(tmpdir):
    destination = str(tmpdir / "convert.json")
    argv = [
        sys.executable,
        "execute",
        "demo",
        "--test",
        "-p",
        "task1:b=42",
        "-o",
        f"convert_destination={destination}",
    ]

    main(argv=argv, shell=False)
    assert os.path.exists(destination)

    graph = load_graph(destination)

    task1_node = graph.graph.nodes["task1"]
    assert has_default_input(task1_node, "b", 42)

    assert graph.graph.graph["requirements"] is not None
    assert _ewokscore_in_graph_requirements(graph)


def test_execute_with_convert_destination_inputs_all(tmpdir):
    destination = str(tmpdir / "convert.json")
    argv = [
        sys.executable,
        "execute",
        "demo",
        "--test",
        "-p",
        "b=42",
        "--inputs=all",
        "-o",
        f"convert_destination={destination}",
    ]

    main(argv=argv, shell=False)
    assert os.path.exists(destination)

    graph = load_graph(destination)

    for node in graph.graph.nodes.values():
        assert has_default_input(node, "b", 42)

    assert graph.graph.graph["requirements"] is not None
    assert _ewokscore_in_graph_requirements(graph)



import json
import subprocess
import pytest


def test_install(venv):
    with pytest.raises(Exception, match="package is not installed"):
        venv.get_version("ewoksdata")

    subprocess.check_call(
        [
            "ewoks",
            "install",
            "--yes",
            '{"graph": {"id": "test_install", "requirements": ["ewoksdata"]}}',
            "-p",
            f"{venv.python}",
        ]
    )

    assert venv.get_version("ewoksdata") is not None


def test_install_with_extract(venv):
    with pytest.raises(Exception, match="package is not installed"):
        venv.get_version("ewoksdata")

    nodes = [
        {
            "id": 1,
            "task_identifier": 'ewoksdata.tasks.normalization.Normalization"',
            "task_type": "class",
        },
        {
            "id": 2,
            "task_identifier": "path/to/my/script",
            "task_type": "script",
        },  # Check that unsupported task type goes though without error
    ]

    graph = {"graph": {"id": "test_install"}, "nodes": nodes}

    subprocess.check_call(
        [
            "ewoks",
            "install",
            "--yes",
            json.dumps(graph),
            "-p",
            f"{venv.python}",
        ]
    )

    assert venv.get_version("ewoksdata") is not None



from . import cli_execute_utils
from . import utils


def add_submit_parameters(parser):
    cli_execute_utils.add_execute_parameters(parser)
    parser.add_argument(
        "--wait",
        type=float,
        default=-1,
        help="Timeout for receiving the result (negative number to disable)",
    )
    parser.add_argument(
        "-c",
        "--cparameter",
        dest="cparameters",
        action="append",
        default=[],
        metavar="NAME=VALUE",
        help="Celery parameters",
    )
    parser.add_argument(
        "--load-remote",
        action="store_true",
        dest="resolve_graph_remotely",
        help="Load the workflow remotely instead of locally",
    )


def apply_submit_parameters(args):
    cli_execute_utils.apply_execute_parameters(args)
    args.cparameters = dict(utils.parse_option(item) for item in args.cparameters)
